// binary operator expression templates

namespace mathlib {

template<class Op, class A, class B> struct matrix_binop {};

template<class A, class B>
struct matrix_binop<binop_mul, A, B> {
	typedef typename select_matrix<A, B>::type possibly_A;
	static const int rows = possibly_A::rows;
	static const int columns = possibly_A::columns;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error;
	static const bool element_op = A::element_op && B::element_op;
	typedef matrix_binop<binop_mul, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int R, int C> float element() const
		{ return a.element<R, C>() * b.element<R, C>(); }
};

// matrix multiplication for x1*1y, x2*2y, x3*3y, x4*4y
// implemented like this to avoid the (admittedly small) overhead suffered by
// introducing a helper class (important operation)
template<class A, class B>
struct matrix_binop<binop_mmul<1>, A, B> {
	static const int rows = A::rows;
	static const int columns = B::columns;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error || A::columns != B::rows;
	static const bool element_op = A::element_op && B::element_op;
	typedef matrix_binop<binop_mmul<1>, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int R, int C> float element() const {
		return a.element<R, 0>() * b.element<0, C>();
	}
};
template<class A, class B>
struct matrix_binop<binop_mmul<2>, A, B> {
	static const int rows = A::rows;
	static const int columns = B::columns;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error || A::columns != B::rows;
	enum { a_eo = A::is_storage || A::element_op && B::columns == 1 };
	enum { b_eo = B::is_storage || B::element_op && A::rows == 1 };
	static const bool element_op = a_eo && b_eo;
	typedef matrix_binop<binop_mmul<2>, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int R, int C> float element() const {
		return a.element<R, 0>() * b.element<0, C>()
			+ a.element<R, 1>() * b.element<1, C>();
	}
};
template<class A, class B>
struct matrix_binop<binop_mmul<3>, A, B> {
	static const int rows = A::rows;
	static const int columns = B::columns;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error || A::columns != B::rows;
	enum { a_eo = A::is_storage || A::element_op && B::columns == 1 };
	enum { b_eo = B::is_storage || B::element_op && A::rows == 1 };
	static const bool element_op = a_eo && b_eo;
	typedef matrix_binop<binop_mmul<3>, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int R, int C> float element() const {
		return a.element<R, 0>() * b.element<0, C>()
			+ a.element<R, 1>() * b.element<1, C>()
			+ a.element<R, 2>() * b.element<2, C>();
	}
};
template<class A, class B>
struct matrix_binop<binop_mmul<4>, A, B> {
	static const int rows = A::rows;
	static const int columns = B::columns;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error || A::columns != B::rows;
	enum { a_eo = A::is_storage || A::element_op && B::columns == 1 };
	enum { b_eo = B::is_storage || B::element_op && A::rows == 1 };
	static const bool element_op = a_eo && b_eo;
	typedef matrix_binop<binop_mmul<4>, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int R, int C> float element() const {
		return a.element<R, 0>() * b.element<0, C>()
			+ a.element<R, 1>() * b.element<1, C>()
			+ a.element<R, 2>() * b.element<2, C>()
			+ a.element<R, 3>() * b.element<3, C>();
	}
};

// scalar product
// same rationale as for mmul
template<class A, class B>
struct matrix_binop<binop_dot<1>, A, B> {
	static const int rows = 1;
	static const int columns = 1;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error || A::rows != 1
		|| B::rows != 1 || A::columns != 1 || B::columns != 1;
	static const bool element_op = A::element_op && B::element_op;
	typedef matrix_binop<binop_dot<1>, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int, int> float element() const {
		return a.element<0, 0>() * b.element<0, 0>();
	}
};
template<class A, class B>
struct matrix_binop<binop_dot<2>, A, B> {
	static const int rows = 1;
	static const int columns = 1;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error || A::rows != 2
		|| B::rows != 2 || A::columns != 1 || B::columns != 1;
	static const bool element_op = A::element_op && B::element_op;
	typedef matrix_binop<binop_dot<1>, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int, int> float element() const {
		return a.element<0, 0>() * b.element<0, 0>()
			+ a.element<1, 0>() * b.element<1, 0>();
	}
};
template<class A, class B>
struct matrix_binop<binop_dot<3>, A, B> {
	static const int rows = 1;
	static const int columns = 1;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error || A::rows != 3
		|| B::rows != 3 || A::columns != 1 || B::columns != 1;
	static const bool element_op = A::element_op && B::element_op;
	typedef matrix_binop<binop_dot<1>, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int, int> float element() const {
		return a.element<0, 0>() * b.element<0, 0>()
			+ a.element<1, 0>() * b.element<1, 0>()
			+ a.element<2, 0>() * b.element<2, 0>();
	}
};
template<class A, class B>
struct matrix_binop<binop_dot<4>, A, B> {
	static const int rows = 1;
	static const int columns = 1;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error || A::rows != 4
		|| B::rows != 4 || A::columns != 1 || B::columns != 1;
	static const bool element_op = A::element_op && B::element_op;
	typedef matrix_binop<binop_dot<1>, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int, int> float element() const {
		return a.element<0, 0>() * b.element<0, 0>()
			+ a.element<1, 0>() * b.element<1, 0>()
			+ a.element<2, 0>() * b.element<2, 0>()
			+ a.element<3, 0>() * b.element<3, 0>();
	}
};

template<class A, class B>
struct matrix_binop<binop_cross, A, B> {
	static const int rows = 3;
	static const int columns = 1;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error
		|| !is_vector_3<A>::value || !is_vector_3<B>::value;
	static const bool element_op = A::is_storage && B::is_storage;
	typedef matrix_binop<binop_cross, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int R, int C> float element() const {
		enum { e1 = (R+1) % 3 };
		enum { e2 = (R+2) % 3 };
		return a.element<e1, 0>() * b.element<e2, 0>()
			- a.element<e2, 0>() * b.element<e1, 0>();
	}
};

template<class A, class B>
struct matrix_binop<binop_div, A, B> {
	typedef typename select_matrix<A, B>::type possibly_A;
	static const int rows = possibly_A::rows;
	static const int columns = possibly_A::columns;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error;
	static const bool element_op = A::element_op && B::element_op;
	typedef matrix_binop<binop_div, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int R, int C> float element() const
		{ return a.element<R, C>() / b.element<R, C>(); }
};

template<class A, class B>
struct matrix_binop<binop_add, A, B> {
	static const int rows = A::rows;
	static const int columns = A::columns;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error || A::rows != B::rows
		|| A::columns != B::columns;
	static const bool element_op = A::element_op && B::element_op;
	typedef matrix_binop<binop_add, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int R, int C> float element() const
		{ return a.element<R, C>() + b.element<R, C>(); }
};

template<class A, class B>
struct matrix_binop<binop_sub, A, B> {
	static const int rows = A::rows;
	static const int columns = B::columns;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error || A::rows != B::rows
		|| A::columns != B::columns;
	static const bool element_op = A::element_op && B::element_op;
	typedef matrix_binop<binop_sub, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
	template<int R, int C> float element() const
		{ return a.element<R, C>() - b.element<R, C>(); }
};

template<class A, class B>
struct matrix_binop<binop_expand, A, B> {
	static const int rows = A::rows + B::rows;
	static const int columns = 1;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error;
	static const bool element_op = false;
	typedef matrix_binop<binop_expand, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
};

template<class A, class B>
struct matrix_binop<binop_rotation, A, B> {
	static const int rows = 4;
	static const int columns = 4;
	static const bool is_storage = false;
	static const bool is_lvalue = false;
	static const bool error = A::error || B::error;
	static const bool element_op = false;
	typedef matrix_binop<binop_rotation, A, B> base_type;
	typename matrix_ref<A>::type a;
	typename matrix_ref<B>::type b;
	matrix_binop(const A &a_, const B &b_) : a(a_), b(b_) {}
};

} // namespace mathlib
